/*
 *	vii - buffer and display outout
 *	Copyright (C) 1991, 1992, 1995, 1997, 1999, 2001, 2005 Peter Miller
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 * MANIFEST: functions to process user commands
 */

#include <ac/stdio.h>
#include <ac/string.h>
#include <ac/time.h>
#include <ac/unistd.h>

#include <error.h>
#include <veneer.h>
#include <stash.h>
#include <str.h>
#include <term.h>

#define CTRL(c) ((c) & 0x9F)

static	int	quitting;
static	int	pos_x;
static	int	pos_y;
	char	*terminal;
	int	sleep_time	= 10;
static	stash	*curpage;
static	stash	*newpage;
static	char	*show_time;
static	int	show_position;


static void update _((void));

static void
update()
{
	int	x, y;
	int	show_row;

	if (!curpage)
		return;
	term_clear();
	term_get_size(&x, &y);
	stash_draw(curpage, pos_x, pos_y, x, y);

	show_row = 0;
	if (show_time)
	{
		char		buffer[80];
		time_t		now;
		struct tm	*tm;

		time(&now);
		tm = localtime(&now);
		buffer[0] = ' ';
		strftime(buffer + 1, sizeof(buffer) - 1, show_time, tm);
		term_puts(x - strlen(buffer), show_row, buffer);
		++show_row;
	}

	if (show_position)
	{
		char		buffer[80];

		sprintf
		(
			buffer,
			" Lines %d-%d of %d",
			pos_y + 1,
			pos_y + y,
			stash_length(curpage)
		);
		term_puts(x - strlen(buffer), show_row, buffer);
		++show_row;
	}

	term_last_line();
	term_refresh();
}


static void remember _((int));

static void
remember(c)
	int	c;
{
	if (!newpage)
		return;
	if (c == '\f')
	{
		if (curpage)
			stash_free(curpage);
		curpage = newpage;
		newpage = stash_open();
		update();
	}
	else
		stash_putc(newpage, c);
}


static void command _((int));

static void
command(c)
	int	c;
{
	int	x, y;
	char	*tmode;

	switch (c)
	{
	default:
		term_bell();
		break;

	case 'q':
	case 'Q':
		quitting = 1;
		break;

	case CTRL('L'):
	case CTRL('R'):
		term_redraw();
		break;

	case CTRL('E'):
	case 'j':
	case '+':
	case '\r':
	case '\n':
		pos_y++;
		update();
		break;

	case CTRL('Y'):
	case 'k':
	case '-':
		if (pos_y > 0)
			pos_y--;
		update();
		break;

	case 'J':
	case ' ':
	case CTRL('F'):
		term_get_size(&x, &y);
		pos_y += y - y / 4;
		update();
		break;

	case 'K':
	case CTRL('B'):
		term_get_size(&x, &y);
		pos_y -= y - y / 4;
		if (pos_y < 0)
			pos_y = 0;
		update();
		break;

	case 'C':
	case 'c':
	case CTRL('C'):
		tmode = "%c"; /* time and date */
		goto set_tmode;

	case 'D':
	case 'd':
	case CTRL('D'):
		tmode = "%x"; /* date w/o time */
		goto set_tmode;

	case 'T':
	case 't':
	case CTRL('T'):
		tmode = "%X"; /* time w/o date */
		set_tmode:
		if (show_time && !strcmp(show_time, tmode))
			show_time = 0;
		else
			show_time = tmode;
		update();
		break;

	case 'P':
	case 'p':
	case CTRL('P'):
		show_position = !show_position;
		update();
		break;
	}
}


static void veneer_gather _((int fd));

static void
veneer_gather(fd)
	int	fd;
{
	int	c;

	newpage = stash_open();
	for(;;)
	{
		c = term_getc(fd, remember);
		if (c < 0)
		{
			if (curpage)
				stash_free(curpage);
			curpage = newpage;
			newpage = 0;
			update();
			break;
		}
		command(c);
		if (quitting)
			break;
	}
}


static void veneer_cmdloop _((void));

static void
veneer_cmdloop()
{
	int	c;

	for(;;)
	{
		c = term_getc(-1, (void (*)_((int)))0);
		if (c < 0)
			break;
		command(c);
		if (quitting)
			break;
	}
}


void
veneer(cmd)
	char	*cmd;
{
	if (cmd)
	{
		term_open(terminal);
		for (;;)
		{
			FILE		*fp;
			time_t		start;
			time_t		finish;
			long		elapsed;
			string_ty	*cmd2;

			cmd2 = str_format("( %s ) 2>&1 < /dev/null", cmd);

			time(&start);
			fp = popen(cmd2->str_text, "r");
			if (!fp)
				nfatal("%s", cmd2->str_text);
			str_free(cmd2);
			veneer_gather(fileno(fp));
			pclose(fp);
			while (!quitting)
			{
				time(&finish);
				elapsed = finish - start;
				if (elapsed >= sleep_time)
					break;
				alarm(sleep_time - elapsed);
				veneer_cmdloop();
				alarm(0);
			}
			if (quitting)
				break;
		}
		term_close();
	}
	else
	{
		int	fd;

		/*
		 * make a copy of the stdin file descriptor,
		 * and reopen stdin to the terminal.
		 */
		fd = dup(0);
		if (!freopen("/dev/tty", "r", stdin))
			nfatal("/dev/tty");

		/*
		 * consume all the input,
		 * and then do commands from stdin
		 */
		term_open(terminal);
		veneer_gather(fd);
		while (!quitting)
			veneer_cmdloop();
		term_close();
	}
}
